/*------------------------------------------------------------------------------------------------------------
RGB-LED Ansteuerung mit 3 PWM-Kanlen

 Realisiert mit ATmega88
 
 Version vom 30.11.2011 
 
 
 Interner RC Oszillator mit 8MHz (Clkdiv8 ausgeschaltet)
 

            (Reset)PC6 1  28 PC5    
             LedT1 PD0 2  27 PC4   
             LedT2 PD1 3  26 PC3   
             LedT3 PD2 4  25 PC2   Poti Sttigung
             LedT4 PD3 5  24 PC1   Poti Farbe
             LedT5 PD4 6  23 PC0   Poti Helligkeit
                   VCC 7  22 GND
                   GND 8  21 AREF
             (Q1)  PB6 9  20 AVCC
             (Q2)  PB7 10 19 PB5   (SCK )
                   PD5 11 18 PB4   (MISO)  
                   PD6 12 17 PB3   (MOSI)  
                   PD7 13 16 PB2   Gruen
             Blau  PB0 14 15 PB1   Rot

             
Die Potis stellen den Beleuchtungseindruck ein. 
Die Leds LedT1..T5 haben gemeinsame Anode. Die Kathoden werden ber die Taster oder den Port nach GND geschaltet.
Drckt man eine der Tasten T2..T5 fr 3 Sekunden, wird die momentane Poti-Einstellung unter der Taste gespeichert 
und die Led unter der Taste flackert kurz zur Besttigung.
Drckt man krzer als 1s, wird die gespeicherte Einstellung benutzt (und die Led unter der Taste leuchtet dauerhaft). 
Taste T1 schaltet auf Poti-Einstellung um.


Timer 1 erzeugt 16Bit PWM fr Rot und Grn (mit 122Hz Bildfrequenz)
Blau wird mit Kombination aus Timer0 und Timer2 angesteuert, wobei Timer2 die unteren, Timer0 die oberen 8 Bit zhlt.

Die Helligkeitswerte werden mittels Tabelle mit exponentieller Kennlinie ausgegeben (Gammawert:2,2).
Die Tabelle wurde mit folgendem Perlscript erzeugt:
	open (OUT,">rgbtab.c")||die;
	print OUT "U16 RGBNormierung[256]={";
	$k=255**2.2;
	$f=65535/$k;
	print "K=$k\n";
	for ($i=0; $i<255; $i++){
	  $y=int(0.9+($i**2.2)*$f);	
	  print OUT "\n" if 0==$i%16;
	  print OUT "$y,";
	  print "$i\t$y\n";
	}
	$y=int(0.9+(255**2.2)*$f);	
	print OUT "$y};";
	close OUT;

  Die unterschiedliche Strahlungsintensitt der LEDs wurde nicht bercksichtigt. 
  Man knnte fr jeden Farbkanal eine angepasste Kennlinie hinterlegen, aber es war mir zuviel Aufwand
  dafr die Meumgebung zu basteln. Ausserdem msste man dann die Helligkeit der strkeren LEDs drosseln, 
  was nicht unbeding erwnscht ist.  Es ist ja nicht besonders schlimm, wenn sich beim Verndern der 
  Farbe die Helligkeit ndert. Man kann anschliessend ja problemlos von Hand die Helligkeit nachjustieren.
  
*/



#define F_CPU 8000000UL	/* CPU clock in Hertz */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <avr/eeprom.h>

#define U8  unsigned char
#define U16 unsigned int
#define U32 unsigned long
#define I16 int
#define I32 long int

#define BIT0 1
#define BIT1 2
#define BIT2 4
#define BIT3 8
#define BIT4 16
#define BIT5 32
#define BIT6 64
#define BIT7 128

#define ROT   OCR1A
#define GRUEN OCR1B
// BLAU  luft nicht mit Hardware-PWM, sondern wird mit Timer1-Overflow eingeschaltet und Timer2-COMPB ausgeschaltet

#define BLAU_EIN {PORTB|=BIT0;}
#define BLAU_AUS {PORTB&=~BIT0;}

#define TASTLED_OFF (PORTD =0)
#define TASTLEDT1 (PORTD = BIT0)
#define TASTLEDT2 (PORTD = BIT1)
#define TASTLEDT3 (PORTD = BIT2)
#define TASTLEDT4 (PORTD = BIT3)
#define TASTLEDT5 (PORTD = BIT4)
#define TASTLEDTP (PORTD = BIT5)

typedef struct{
  U16 messwert[3];
  U16 gemerktePotistellung[3];
  U8  holeWerte;
  U8  r,g,b;
  U8  pwmR, pwmG, pwmB;
  U8  blauH;
  U8  blauL;
  U8  modus; //0=aus 1=0..255  2=256..65534  3=65535
  U8  timer1overflow;
}Tglob;

volatile Tglob  g;

U16 RGBNormierung[256]={//Helligkeitswerte mit exponentieller Kennlinie 
  0,1,2,4,7,12,18,24,33,42,53,65,79,94,111,129,
  149,170,193,217,243,270,299,330,362,396,432,469,508,549,592,636,
  682,730,779,830,883,938,995,1053,1114,1176,1240,1306,1373,1443,1514,1588,
  1663,1740,1819,1900,1983,2068,2155,2244,2334,2427,2522,2618,2717,2818,2920,3025,
  3131,3240,3351,3463,3578,3695,3814,3935,4058,4183,4310,4439,4570,4703,4839,4976,
  5116,5258,5401,5547,5695,5846,5998,6153,6309,6468,6629,6792,6957,7125,7294,7466,
  7640,7816,7995,8175,8358,8543,8730,8920,9111,9305,9501,9699,9900,10103,10308,10515,
  10725,10936,11150,11367,11585,11806,12029,12255,12482,12712,12945,13179,13416,13655,13897,14141,
  14387,14635,14886,15139,15394,15652,15912,16174,16439,16706,16976,17248,17522,17798,18077,18358,
  18642,18928,19216,19507,19800,20096,20394,20694,20997,21302,21609,21919,22231,22546,22863,23183,
  23505,23829,24156,24485,24817,25151,25488,25827,26168,26512,26858,27207,27558,27912,28268,28627,
  28988,29352,29718,30086,30457,30831,31207,31585,31966,32350,32735,33124,33515,33908,34304,34703,
  35104,35507,35913,36322,36733,37146,37562,37981,38402,38826,39252,39681,40112,40546,40982,41421,
  41863,42307,42753,43203,43654,44109,44566,45025,45487,45952,46419,46888,47361,47836,48313,48793,
  49276,49761,50249,50739,51232,51728,52226,52727,53231,53737,54245,54757,55271,55787,56306,56828,
  57353,57880,58409,58942,59477,60014,60554,61097,61643,62191,62742,63295,63851,64410,64971,65535
};



//--------------------------------------------------------
void setBlue(U16 value){
  if (0==value) g.modus=0; else
  if (value<256) g.modus=1; else
  if (value<65535) g.modus=2; else
  g.modus=3;
  g.blauH = value >>8;
  g.blauL = value & 0xff;
}


//-------------------------------------------------------- 
ISR(ADC_vect){ //Potistellungen messen
//-------------------------------------------------------- 
static U8 n=0;
static U8 kanal=0;
static U16 wert[3]={0,0,0};
static U16 sum[3]={0,0,0};
U8 i;

  if (n>0) sum[kanal]+=ADC; //Messung nach Kanalwechsel auslassen
  n++;
  if (n>32){ 
    n=0;
    kanal++;
    if (kanal > 2){
      kanal=0;
      if (g.holeWerte){
        for (i=0; i<3; i++){
          if ( sum[i] >= wert[i]){
            if ((sum[i] - wert[i]) > 64){
              wert[i] = sum[i];
              g.messwert[i] = wert[i]>>7;  //32 Messungen mit 10 Bit ergeben 15 Bit. 15Bit-7Bit=8Bit
            }
          }else{
            if ((- sum[i] + wert[i]) > 64){
              wert[i] = sum[i];
              g.messwert[i] = wert[i]>>7;  //Messwert enthaelt nun Wert von 0..255
            }
          }
        }
        g.holeWerte=0;
      }
      sum[0]=0;
      sum[1]=0;
      sum[2]=0;
    }
    ADMUX = kanal|(1<<REFS0);     //  Uref:AVCC
  }
  ADCSRA |=(1<<ADSC);
}

//--------------------------------------------------------
void setHSV(U8 hue, U8 sat, U8 val){//RGB-Leds mit HSV-Werten einstellen  Farbe0..251, Sttigung0..255, Helligkeit0..255
//--------------------------------------------------------

  U16 subhue;
  U32 y1,y2,y3;

  //Farbkreis aufteilen in 6 Abschnitte (1/6 Farbkreis: Pizzastck)
  if (0==sat){
    g.r=val;
    g.g=val;
    g.b=val;
  }else{
    if (hue>251) hue=251;
    subhue = (U16)(hue%42) * 255 / 41; //Pizzastck 0..255
    hue /= 42; //0..5
/*
    f= subhue 
    y1 = v * (1 - s);
    y2 = v * (1 - (s * f));
    y3 = v * (1 - (s * (1 - f)));
*/

    y1 = (val * (255 - sat))/255;
    y2 = (val * ( 255 - (sat * subhue)/255 ))/255;
    y3 = (val * ( 255 - (sat * ( 255 - subhue ))/255))/255;
    switch( hue ) {
      case 0:
              g.r = val;
              g.g = y3;
              g.b = y1;
              break;
      case 1:
              g.r = y2;
              g.g = val;
              g.b = y1;
              break;
      case 2:
              g.r = y1;
              g.g = val;
              g.b = y3;
              break;
      case 3:
              g.r = y1;
              g.g = y2;
              g.b = val;
              break;
      case 4:
              g.r = y3;
              g.g = y1;
              g.b = val;
              break;
      default:
              g.r = val;
              g.g = y1;
              g.b = y2;
              break;
    }//switch      
  }//else 
  
  //PWM setzen
  ROT  =RGBNormierung[g.r];
  GRUEN=RGBNormierung[g.g]; 
  setBlue(RGBNormierung[g.b]);
}
//--------------------------------------------------------
void demoLightshow(void){
  U8 hue=0;
  U8 sat=250;
  U8 val = 255;

  //Eine Runde im Farbkreis	
  for (hue=0; hue<252; hue++){
    setHSV(hue, sat, val); //auf RGB-PWM-Ausgnge
    _delay_ms(12);
  }
}
//--------------------------------------------------------
void demoSingle(void){ //Einzelne Farben testen
  U8 val = 128;

   GRUEN=0;
   ROT=0;
#if 1
   for (val=50; val>0; val--){
     ROT=RGBNormierung[val];
     GRUEN=RGBNormierung[1];
     _delay_ms(300);
   }
#endif
#if 1
   ROT=0;
   for (val=0; val<254; val++){
     GRUEN=RGBNormierung[val];
     _delay_ms(150);
   }
#endif
#if 1
   ROT=1;
   for (val=0; val<254; val++){
     setBlue(RGBNormierung[val]);
     _delay_ms(120);
   }
#endif
#if 0
   for (val=255; val>0; val--){
     ROT=RGBNormierung[val];
     _delay_ms(100);
   }
#endif
#if 0
   for (val=255; val>0; val--){
     GRUEN=RGBNormierung[val];
     _delay_ms(150);
   }
#endif

#if 0
   for (val=254; val>0; val--){
     setBlue(RGBNormierung[val]);
     _delay_ms(120);
   }
#endif
}


//--------------------------------------------------------
ISR(TIMER1_OVF_vect){ //alle 8192s, ca. 122Hz (=Bildfrequenz)
//--------------------------------------------------------
  OCR0A = g.blauH;
  OCR2A = g.blauL;
  switch(g.modus){
  case 0:  //0
    BLAU_AUS; 
    TCCR0B = 0; 
    TCCR2B = 0;
    TCNT0=0; 
    TCNT2=0; 
    break;
  case 1: //bis 255
    BLAU_EIN;
    TCCR0B = 0; 
    TCCR2B = (1<<CS20);  // Clk/1
    break;
  case 2: //bis 65534
    BLAU_EIN;
    TCCR0B = (1<<CS02);  //Clk/256
    break;
  default: //65535
    BLAU_EIN; 
    TCCR0B = 0; 
    TCCR2B = 0;
    break;
  }
  g.timer1overflow=1;
}
//--------------------------------------------------------
ISR(TIMER0_COMPA_vect){
//--------------------------------------------------------
  //Starte Timer 2
  TCNT2=0;
  TCCR2B = (1<<CS20);  // Clk/1
 }
//--------------------------------------------------------
ISR(TIMER2_COMPB_vect){
//--------------------------------------------------------
  BLAU_AUS;
  TCCR2B = 0; 
  TCNT2=0;
}
//--------------------------------------------------------
void write_eeprom_value (U8 r, U8 g, U8 b, U8 slot){
//--------------------------------------------------------
  U16 eeprom_adresse;

  eeprom_adresse = slot * 3;
  
  //Werte ins EEPROM schreiben
  eeprom_busy_wait ();
  eeprom_write_byte((U8 *)(eeprom_adresse + 0),r);
  eeprom_busy_wait ();
  eeprom_write_byte((U8 *)(eeprom_adresse + 1),g);
  eeprom_busy_wait ();
  eeprom_write_byte((U8 *)(eeprom_adresse + 2),b);
}

//--------------------------------------------------------
void get_eeprom_value (U8 volatile *r, U8 volatile *g, U8 volatile *b, U8 slot){
//--------------------------------------------------------
  U16 eeprom_adresse;
  eeprom_adresse = slot * 3;
	
  //eeprom_busy_wait ();	
  *r = eeprom_read_byte((U8 *)(eeprom_adresse + 0));
  *g = eeprom_read_byte((U8 *)(eeprom_adresse + 1));
  *b = eeprom_read_byte((U8 *)(eeprom_adresse + 2));
}

//--------------------------------------------------------
void tastenabfrage(void){ //Aufruf im Mittel alle 10ms
//--------------------------------------------------------
//Tasten abfragen, entprellen und bewerten

U8 taste;
static U8 tasteBisher=0;
static U16 timer_ms=0;
U8 i;
U8 Ledstatus;	
U8 Tastenstatus;	
U8 mask = BIT0 | BIT1 | BIT2 | BIT3 | BIT4;	
	
  Ledstatus = PORTD; //Merke Ledstatus	
  DDRD = 0;          //Leds aus und auf Eingang schalten
  PORTD = mask;      //Pull-Ups einschalten
  asm("NOP");        //Kurz warten	
  Tastenstatus = ~(PIND & mask); //Tastenstatus lesen (negieren, da Taster nach GND schalten)	
  DDRD = mask;       //Ports wieder auf Ausgang
  PORTD = Ledstatus; //Alten Status wiederherstellen	
	
  taste=0;
  if (Tastenstatus > 0){ //Eine Taste gedrckt
    PORTD = Tastenstatus;   //entsprechende LED einschalten 	  
    if (Tastenstatus & BIT0) taste=1;
    if (Tastenstatus & BIT1) taste=2;
    if (Tastenstatus & BIT2) taste=3;
    if (Tastenstatus & BIT3) taste=4;
    if (Tastenstatus & BIT4) taste=5;
  }	

  // Umschalten bei Loslassen nach kurzer Zeit

  if ((timer_ms >= 50) && (timer_ms < 1200)){
    if ((0==taste) && (tasteBisher > 1)){
      //Umschalten auf gespeicherte Einstellung
      get_eeprom_value (&g.r, &g.g, &g.b, tasteBisher);
      ROT  =RGBNormierung[g.r];
      GRUEN=RGBNormierung[g.g]; 
      setBlue(RGBNormierung[g.b]);
      g. gemerktePotistellung[0] = g.messwert[0];
      g. gemerktePotistellung[1] = g.messwert[1];
      g. gemerktePotistellung[2] = g.messwert[2];
    }
  }else
  if ((timer_ms > 2200) && (0!=taste) && (timer_ms<3000) ){ //Speichern bei langem gedrckt halten
    g. gemerktePotistellung[0] = g.messwert[0];
    g. gemerktePotistellung[1] = g.messwert[1];
    g. gemerktePotistellung[2] = g.messwert[2];
    //Durch blinken besttigen
    for (i=0; i<5; i++){
      PORTD = 0;   //LEDs aus
      _delay_ms(60);
      PORTD = Tastenstatus;   //entsprechende LED einschalten 	  
      _delay_ms(40);
    }
    // Einstellung speichern
    write_eeprom_value (g.r, g.g, g.b, tasteBisher);
    timer_ms = 4000; //Blockiere mehrfaches speichern
  }
  if (timer_ms < 4000)  timer_ms+=10;

  if (tasteBisher!=taste){
    tasteBisher=taste;
    timer_ms=0;
  }
}



//--------------------------------------------------------
int main(void){
//--------------------------------------------------------
  U16 time_us=0;
  cli();
  DDRB = BIT0 | BIT1 | BIT2; //RGB-Leds
  DDRC = 0;
  DDRD = BIT0 | BIT1 | BIT2 | BIT3 | BIT4; //Tasten-LEDs

  PORTB=0;
  PORTC=BIT3 | BIT4 | BIT5; //Pull-Ups fr Tasten
  PORTD=BIT0; //Taste1: Potis aktiv

  //Timer 0 fr Blau (High Byte)
  TCCR0A = (1<<WGM01)|(1<<WGM00); //Fast PWM
  TCCR0B = (1<<CS02);  // Clk/256
  OCR0A = 1;
  TIMSK0 =(1<<OCIE0A);

  //Timer 1 als 16-Bit PWM (Mode 14)  fr Rot+Grn
  TCCR1A = (1<<COM1A1)|(1<<COM1B1)/*|(1<<COM1A0)|(1<<COM1B0)*/|(1<<WGM11);
  TCCR1B = (1<<WGM13)|(1<<WGM12)|(1<<CS10);  // Clk/1  ==>122 Hz PWM (alle 8,192 ms)
  OCR1A = 30000;
  OCR1B = 30000;
  ICR1=65535;
  TIMSK1 = (1<<TOIE1);

  //Timer 2 fr Blau (LowByte)
  TCCR2A = (1<<COM2B1)|(1<<COM2B0); //Normal (Mode 0)
  TCCR2B = (1<<CS20);  // Clk/1  ==>488 Hz PWM
  OCR2A = 255;
  OCR2B = 1;
  TIMSK2 = (1<<OCIE2B);

  sei(); //Interrupts scharf

//ADC:
  ADMUX  = (1<<REFS0);     //  Uref=Versorgungsspannung
  ADCSRA = (1<<ADEN) | (1<<ADSC) | (1<<ADATE) | (1<<ADIF) | (1<<ADIE)     
         | (1<<ADPS2)| (1<<ADPS1)| (1<<ADPS0); //8MHZ / 128 = 62500Hz
  ADCSRB = 0;  //Free running
  DIDR0 = BIT0 | BIT1 | BIT2;



  while(1){  demoSingle();  }  
//  while(1){  demoLightshow();}
  demoLightshow();
  TASTLEDTP;
  while(1){
    if (g.timer1overflow){
      g.timer1overflow--;
      time_us+=8192;
      if (time_us >= 10000){
        time_us-=10000;
        tastenabfrage(); //im Mittel alle 10 ms
        if (!g.holeWerte){ //ADC hat neue Werte geliefert
          if (1==PORTD){  //Potis aktiv
            setHSV(g.messwert[0],g.messwert[1],g.messwert[2]); //auf RGB-PWM-Ausgnge
          }
          g.holeWerte=1;
        }
      }
    }
  }//While 1
  return 0;
}
//EOF
